"""
Ticker Stock Schemas:
    dividends
    splits
    actions
    capital_gains
    shares_full
    info
    fast_info
    news
"""

from datetime import date, datetime
from typing import Any

import pandas as pd
from pydantic import BaseModel, Field, field_validator

from openmarkets.schemas.company import CompanyOfficer


class StockFastInfo(BaseModel):
    """Fast info snapshot for a stock ticker, typically from yfinance or similar APIs."""

    currency: str = Field(..., description="Currency of the ticker.")
    dayHigh: float = Field(..., description="Day's high price.")
    dayLow: float = Field(..., description="Day's low price.")
    exchange: str = Field(..., description="Exchange where the ticker is listed.")
    fiftyDayAverage: float = Field(..., description="50-day average price.")
    lastPrice: float = Field(..., description="Last traded price.")
    lastVolume: int = Field(..., description="Last traded volume.")
    marketCap: float | None = Field(None, description="Market capitalization.")
    open: float = Field(..., description="Opening price.")
    previousClose: float = Field(..., description="Previous closing price.")
    quoteType: str = Field(..., description="Type of quote (e.g., equity, ETF).")
    regularMarketPreviousClose: float = Field(..., description="Regular market previous close.")
    shares: int | None = Field(None, description="Number of shares outstanding.")
    tenDayAverageVolume: int = Field(..., description="10-day average volume.")
    threeMonthAverageVolume: int = Field(..., description="3-month average volume.")
    timezone: str = Field(..., description="Timezone of the exchange.")
    twoHundredDayAverage: float = Field(..., description="200-day average price.")
    yearChange: float = Field(..., description="Change over the past year.")
    yearHigh: float = Field(..., description="52-week high price.")
    yearLow: float = Field(..., description="52-week low price.")


class StockInfo(BaseModel):
    """
    Comprehensive schema for stock information, typically from yfinance Ticker.info.
    """

    address1: str | None = Field(None, description="Primary address line of the company.")
    city: str | None = Field(None, description="City of the company's headquarters.")
    state: str | None = Field(None, description="State or province of the company's headquarters.")
    zip: str | None = Field(None, description="Postal code of the company's headquarters.")
    country: str | None = Field(None, description="Country of the company's headquarters.")
    phone: str | None = Field(None, description="Contact phone number.")
    website: str | None = Field(None, description="Company website URL.")
    industry: str | None = Field(None, description="Industry classification.")
    industryKey: str | None = Field(None, description="Industry key.")
    industryDisp: str | None = Field(None, description="Industry display name.")
    sector: str | None = Field(None, description="Sector classification.")
    sectorKey: str | None = Field(None, description="Sector key.")
    sectorDisp: str | None = Field(None, description="Sector display name.")
    longBusinessSummary: str | None = Field(None, description="Extended business description.")
    fullTimeEmployees: int | None = Field(None, description="Number of full-time employees.")
    companyOfficers: list[CompanyOfficer] | None = Field(None, description="List of company officers.")
    auditRisk: int | None = Field(None, description="Audit risk score.")
    boardRisk: int | None = Field(None, description="Board risk score.")
    compensationRisk: int | None = Field(None, description="Compensation risk score.")
    shareHolderRightsRisk: int | None = Field(None, description="Shareholder rights risk score.")
    overallRisk: int | None = Field(None, description="Overall risk score.")
    governanceEpochDate: int | None = Field(None, description="Governance epoch date (timestamp).")
    compensationAsOfEpochDate: int | None = Field(None, description="Compensation as of epoch date (timestamp).")
    irWebsite: str | None = Field(None, description="Investor relations website.")
    executiveTeam: list[Any] | None = Field(None, description="List of executive team members.")
    maxAge: int | None = Field(None, description="Maximum age of the data.")
    priceHint: int | None = Field(None, description="Price hint for display.")
    previousClose: float | None = Field(None, description="Previous closing price.")
    open: float | None = Field(None, description="Opening price.")
    dayLow: float | None = Field(None, description="Day's low price.")
    dayHigh: float | None = Field(None, description="Day's high price.")
    regularMarketPreviousClose: float | None = Field(None, description="Regular market previous close.")
    regularMarketOpen: float | None = Field(None, description="Regular market opening price.")
    regularMarketDayLow: float | None = Field(None, description="Regular market day's low price.")
    regularMarketDayHigh: float | None = Field(None, description="Regular market day's high price.")
    dividendRate: float | None = Field(None, description="Dividend rate.")
    dividendYield: float | None = Field(None, description="Dividend yield.")
    exDividendDate: datetime | None = Field(None, description="Ex-dividend date.")
    payoutRatio: float | None = Field(None, description="Payout ratio.")
    fiveYearAvgDividendYield: float | None = Field(None, description="Five-year average dividend yield.")
    beta: float | None = Field(None, description="Beta value.")
    trailingPE: float | None = Field(None, description="Trailing P/E ratio.")
    forwardPE: float | None = Field(None, description="Forward P/E ratio.")
    volume: int | None = Field(None, description="Trading volume.")
    regularMarketVolume: int | None = Field(None, description="Regular market trading volume.")
    averageVolume: int | None = Field(None, description="Average trading volume.")
    averageVolume10days: int | None = Field(None, description="10-day average trading volume.")
    averageDailyVolume10Day: int | None = Field(None, description="10-day average daily volume.")
    bid: float | None = Field(None, description="Bid price.")
    ask: float | None = Field(None, description="Ask price.")
    bidSize: int | None = Field(None, description="Bid size.")
    askSize: int | None = Field(None, description="Ask size.")
    marketCap: int | None = Field(None, description="Market capitalization.")
    fiftyTwoWeekLow: float | None = Field(None, description="52-week low price.")
    fiftyTwoWeekHigh: float | None = Field(None, description="52-week high price.")
    allTimeHigh: float | None = Field(None, description="All-time high price.")
    allTimeLow: float | None = Field(None, description="All-time low price.")
    priceToSalesTrailing12Months: float | None = Field(None, description="Price to sales (TTM).")
    fiftyDayAverage: float | None = Field(None, description="50-day average price.")
    twoHundredDayAverage: float | None = Field(None, description="200-day average price.")
    trailingAnnualDividendRate: float | None = Field(None, description="Trailing annual dividend rate.")
    trailingAnnualDividendYield: float | None = Field(None, description="Trailing annual dividend yield.")
    currency: str | None = Field(None, description="Trading currency.")
    tradeable: bool | None = Field(None, description="Is the stock tradeable.")
    enterpriseValue: int | None = Field(None, description="Enterprise value.")
    profitMargins: float | None = Field(None, description="Profit margins.")
    floatShares: int | None = Field(None, description="Float shares.")
    sharesOutstanding: int | None = Field(None, description="Shares outstanding.")
    sharesShort: int | None = Field(None, description="Shares short.")
    sharesShortPriorMonth: int | None = Field(None, description="Shares short prior month.")
    sharesShortPreviousMonthDate: int | None = Field(None, description="Shares short previous month date.")
    dateShortInterest: int | None = Field(None, description="Date of short interest.")
    sharesPercentSharesOut: float | None = Field(None, description="Percent shares out.")
    heldPercentInsiders: float | None = Field(None, description="Percent held by insiders.")
    heldPercentInstitutions: float | None = Field(None, description="Percent held by institutions.")
    shortRatio: float | None = Field(None, description="Short ratio.")
    shortPercentOfFloat: float | None = Field(None, description="Short percent of float.")
    impliedSharesOutstanding: int | None = Field(None, description="Implied shares outstanding.")
    bookValue: float | None = Field(None, description="Book value.")
    priceToBook: float | None = Field(None, description="Price to book ratio.")
    lastFiscalYearEnd: datetime | None = Field(None, description="Last fiscal year end.")
    nextFiscalYearEnd: datetime | None = Field(None, description="Next fiscal year end.")
    mostRecentQuarter: int | None = Field(None, description="Most recent quarter (timestamp).")
    earningsQuarterlyGrowth: float | None = Field(None, description="Earnings quarterly growth.")
    netIncomeToCommon: int | None = Field(None, description="Net income to common.")
    trailingEps: float | None = Field(None, description="Trailing EPS.")
    forwardEps: float | None = Field(None, description="Forward EPS.")
    lastSplitFactor: str | None = Field(None, description="Last split factor.")
    lastSplitDate: datetime | None = Field(None, description="Last split date.")
    enterpriseToRevenue: float | None = Field(None, description="Enterprise to revenue ratio.")
    enterpriseToEbitda: float | None = Field(None, description="Enterprise to EBITDA ratio.")
    FiftyTwoWeekChange: float | None = Field(None, alias="52WeekChange", description="52-week change.")
    SandP52WeekChange: float | None = Field(None, description="S&P 52-week change.")
    lastDividendValue: float | None = Field(None, description="Last dividend value.")
    lastDividendDate: datetime | None = Field(None, description="Last dividend date.")
    quoteType: str | None = Field(None, description="Quote type.")
    currentPrice: float | None = Field(None, description="Current price.")
    targetHighPrice: float | None = Field(None, description="Target high price.")
    targetLowPrice: float | None = Field(None, description="Target low price.")
    targetMeanPrice: float | None = Field(None, description="Target mean price.")
    targetMedianPrice: float | None = Field(None, description="Target median price.")
    recommendationMean: float | None = Field(None, description="Recommendation mean.")
    recommendationKey: str | None = Field(None, description="Recommendation key.")
    numberOfAnalystOpinions: int | None = Field(None, description="Number of analyst opinions.")
    totalCash: int | None = Field(None, description="Total cash.")
    totalCashPerShare: float | None = Field(None, description="Total cash per share.")
    ebitda: int | None = Field(None, description="EBITDA.")
    totalDebt: int | None = Field(None, description="Total debt.")
    quickRatio: float | None = Field(None, description="Quick ratio.")
    currentRatio: float | None = Field(None, description="Current ratio.")
    totalRevenue: int | None = Field(None, description="Total revenue.")
    debtToEquity: float | None = Field(None, description="Debt to equity ratio.")
    revenuePerShare: float | None = Field(None, description="Revenue per share.")
    returnOnAssets: float | None = Field(None, description="Return on assets.")
    returnOnEquity: float | None = Field(None, description="Return on equity.")
    grossProfits: int | None = Field(None, description="Gross profits.")
    freeCashflow: int | None = Field(None, description="Free cash flow.")
    operatingCashflow: int | None = Field(None, description="Operating cash flow.")
    earningsGrowth: float | None = Field(None, description="Earnings growth.")
    revenueGrowth: float | None = Field(None, description="Revenue growth.")
    grossMargins: float | None = Field(None, description="Gross margins.")
    ebitdaMargins: float | None = Field(None, description="EBITDA margins.")
    operatingMargins: float | None = Field(None, description="Operating margins.")
    financialCurrency: str | None = Field(None, description="Financial reporting currency.")
    symbol: str | None = Field(None, description="Ticker symbol.")
    language: str | None = Field(None, description="Reporting language.")
    region: str | None = Field(None, description="Region.")
    typeDisp: str | None = Field(None, description="Type display name.")
    quoteSourceName: str | None = Field(None, description="Quote source name.")
    triggerable: bool | None = Field(None, description="Is triggerable.")
    customPriceAlertConfidence: str | None = Field(None, description="Custom price alert confidence.")
    regularMarketChangePercent: float | None = Field(None, description="Regular market change percent.")
    regularMarketPrice: float | None = Field(None, description="Regular market price.")
    shortName: str | None = Field(None, description="Short name.")
    longName: str | None = Field(None, description="Long name.")
    hasPrePostMarketData: bool | None = Field(None, description="Has pre/post market data.")
    firstTradeDateMilliseconds: int | None = Field(None, description="First trade date in milliseconds.")
    postMarketChangePercent: float | None = Field(None, description="Post-market change percent.")
    postMarketPrice: float | None = Field(None, description="Post-market price.")
    postMarketChange: float | None = Field(None, description="Post-market change.")
    regularMarketChange: float | None = Field(None, description="Regular market change.")
    regularMarketDayRange: str | None = Field(None, description="Regular market day range.")
    fullExchangeName: str | None = Field(None, description="Full exchange name.")
    averageDailyVolume3Month: int | None = Field(None, description="3-month average daily volume.")
    fiftyTwoWeekLowChange: float | None = Field(None, description="52-week low change.")
    fiftyTwoWeekLowChangePercent: float | None = Field(None, description="52-week low change percent.")
    fiftyTwoWeekRange: str | None = Field(None, description="52-week range.")
    fiftyTwoWeekHighChange: float | None = Field(None, description="52-week high change.")
    fiftyTwoWeekHighChangePercent: float | None = Field(None, description="52-week high change percent.")
    fiftyTwoWeekChangePercent: float | None = Field(None, description="52-week change percent.")
    marketState: str | None = Field(None, description="Market state.")
    corporateActions: list[Any] | None = Field(None, description="Corporate actions.")
    postMarketTime: int | None = Field(None, description="Post-market time (timestamp).")
    regularMarketTime: int | None = Field(None, description="Regular market time (timestamp).")
    exchange: str | None = Field(None, description="Exchange code.")
    messageBoardId: str | None = Field(None, description="Message board ID.")
    exchangeTimezoneName: str | None = Field(None, description="Exchange timezone name.")
    exchangeTimezoneShortName: str | None = Field(None, description="Exchange timezone short name.")
    gmtOffSetMilliseconds: int | None = Field(None, description="GMT offset in milliseconds.")
    market: str | None = Field(None, description="Market name.")
    esgPopulated: bool | None = Field(None, description="ESG data populated.")
    dividendDate: datetime | None = Field(None, description="Dividend date.")
    earningsTimestamp: datetime | None = Field(None, description="Earnings timestamp.")
    earningsTimestampStart: datetime | None = Field(None, description="Earnings timestamp start.")
    earningsTimestampEnd: datetime | None = Field(None, description="Earnings timestamp end.")
    earningsCallTimestampStart: datetime | None = Field(None, description="Earnings call timestamp start.")
    earningsCallTimestampEnd: datetime | None = Field(None, description="Earnings call timestamp end.")
    isEarningsDateEstimate: bool | None = Field(None, description="Is earnings date an estimate.")
    epsTrailingTwelveMonths: float | None = Field(None, description="EPS trailing twelve months.")
    epsForward: float | None = Field(None, description="EPS forward.")
    epsCurrentYear: float | None = Field(None, description="EPS current year.")
    priceEpsCurrentYear: float | None = Field(None, description="Price/EPS current year.")
    fiftyDayAverageChange: float | None = Field(None, description="50-day average change.")
    fiftyDayAverageChangePercent: float | None = Field(None, description="50-day average change percent.")
    twoHundredDayAverageChange: float | None = Field(None, description="200-day average change.")
    twoHundredDayAverageChangePercent: float | None = Field(None, description="200-day average change percent.")
    sourceInterval: int | None = Field(None, description="Source interval.")
    exchangeDataDelayedBy: int | None = Field(None, description="Exchange data delayed by (seconds).")
    averageAnalystRating: str | None = Field(None, description="Average analyst rating.")
    cryptoTradeable: bool | None = Field(None, description="Is crypto tradeable.")
    displayName: str | None = Field(None, description="Display name.")
    trailingPegRatio: float | None = Field(None, description="Trailing PEG ratio.")

    @field_validator(
        "exDividendDate",
        "lastDividendDate",
        "dividendDate",
        "lastSplitDate",
        "earningsTimestamp",
        "earningsTimestampStart",
        "earningsTimestampEnd",
        "earningsCallTimestampStart",
        "earningsCallTimestampEnd",
        "lastFiscalYearEnd",
        "nextFiscalYearEnd",
        mode="before",
    )
    @classmethod
    def _convert_to_datetime(cls, v):
        """Convert Unix timestamp (int/str) to datetime, or pass through if already datetime/None."""
        if v is None or isinstance(v, datetime):
            return v
        try:
            ts = int(float(v))
            return datetime.fromtimestamp(ts)
        except Exception:
            return None


class StockDividends(BaseModel):
    """Dividend payment for a ticker."""

    date_: datetime = Field(..., description="Date of the dividend payment.", alias="Date")
    dividend: float = Field(..., description="Dividend amount.", alias="Dividends")

    @field_validator("date_")
    def parse_date(cls, value) -> date:
        """Validator to parse date from timestamp to date."""
        if isinstance(value, pd.Timestamp):
            return value.to_pydatetime()
        return value


class StockHistory(BaseModel):
    """Schema for historical ticker data (OHLCV, splits, dividends)."""

    Date: datetime = Field(..., description="Date of record")
    Open: float = Field(..., description="Opening price")
    High: float = Field(..., description="Highest price")
    Low: float = Field(..., description="Lowest price")
    Close: float = Field(..., description="Closing price")
    Volume: int = Field(..., description="Volume traded")
    Dividends: float | None = Field(None, description="Dividends paid")
    Stock_Splits: float | None = Field(None, alias="Stock Splits", description="Stock splits")


class StockInfo_v2(BaseModel):
    """Schema for general stock information."""

    symbol: str | None = Field(None, description="Ticker symbol")
    shortName: str | None = Field(None, description="Short name of the company")
    longName: str | None = Field(None, description="Long name of the company")
    sector: str | None = Field(None, description="Sector of the company")
    industry: str | None = Field(None, description="Industry of the company")
    marketCap: int | None = Field(None, description="Market capitalization")
    currentPrice: float | None = Field(None, description="Current trading price")
    previousClose: float | None = Field(None, description="Previous closing price")
    open: float | None = Field(None, description="Opening price")
    dayLow: float | None = Field(None, description="Lowest price of the day")
    dayHigh: float | None = Field(None, description="Highest price of the day")
    volume: int | None = Field(None, description="Trading volume")
    averageVolume: int | None = Field(None, description="Average trading volume")
    beta: float | None = Field(None, description="Beta value")
    trailingPE: float | None = Field(None, description="Trailing P/E ratio")
    forwardPE: float | None = Field(None, description="Forward P/E ratio")
    dividendYield: float | None = Field(None, description="Dividend yield")
    payoutRatio: float | None = Field(None, description="Payout ratio")
    fiftyTwoWeekLow: float | None = Field(None, description="52-week low price")
    fiftyTwoWeekHigh: float | None = Field(None, description="52-week high price")
    priceToBook: float | None = Field(None, description="Price to book ratio")
    debtToEquity: float | None = Field(None, description="Debt to equity ratio")
    returnOnEquity: float | None = Field(None, description="Return on equity")
    returnOnAssets: float | None = Field(None, description="Return on assets")
    freeCashflow: float | None = Field(None, description="Free cash flow")
    operatingCashflow: float | None = Field(None, description="Operating cash flow")
    website: str | None = Field(None, description="Company website")
    country: str | None = Field(None, description="Country of headquarters")
    city: str | None = Field(None, description="City of headquarters")
    phone: str | None = Field(None, description="Contact phone number")
    fullTimeEmployees: int | None = Field(None, description="Number of full-time employees")
    longBusinessSummary: str | None = Field(None, description="Long business summary")
    exDividendDate: datetime | None = Field(None, description="Ex-dividend date as datetime")

    @field_validator("exDividendDate", mode="before")
    @classmethod
    def convert_ex_dividend_date(cls, v):
        """Convert exDividendDate from Unix timestamp (int/str) to datetime, or pass through if already datetime/None."""
        if v is None or isinstance(v, datetime):
            return v
        try:
            # Accept int, float, or string representations of Unix timestamp
            ts = int(float(v))
            return datetime.fromtimestamp(ts)
        except Exception:
            return None


class StockSplit(BaseModel):
    """Stock split event for a ticker."""

    date: datetime = Field(..., description="Date of the stock split.")
    stock_splits: float = Field(..., description="Stock split")


class CorporateActions(BaseModel):
    """Actions for a ticker."""

    date: datetime = Field(..., description="Date of the action.", alias="Date")
    dividend: float | None = Field(None, description="Dividend amount.", alias="Dividends")
    stock_splits: float | None = Field(None, description="Stock split amount.", alias="Stock Splits")


class NewsItem(BaseModel):
    """News item for a stock ticker."""

    id: str = Field(..., description="Unique identifier for the news item.")
    content: dict = Field(..., description="Content of the news item.")



"""Analyst data tools."""

import inspect
import json
import sys

import yfinance as yf
from mcp.server import FastMCP


async def get_recommendations(ticker: str) -> str:
    """Get analyst recommendations for a stock.

    Args:
        ticker: Stock ticker symbol

    Returns:
        JSON string containing analyst recommendations
    """
    stock = yf.Ticker(ticker)
    recs = stock.recommendations
    if recs is not None and not recs.empty:
        return recs.to_json(date_format="iso")
    return json.dumps({"error": "No recommendations available"})


async def get_analyst_price_targets(ticker: str) -> str:
    """Get analyst price targets and estimates.

    Args:
        ticker: Stock ticker symbol

    Returns:
        JSON string containing price targets
    """
    stock = yf.Ticker(ticker)
    info = stock.info
    price_targets = {
        "targetHighPrice": info.get("targetHighPrice"),
        "targetLowPrice": info.get("targetLowPrice"),
        "targetMeanPrice": info.get("targetMeanPrice"),
        "targetMedianPrice": info.get("targetMedianPrice"),
        "recommendationMean": info.get("recommendationMean"),
        "recommendationKey": info.get("recommendationKey"),
        "numberOfAnalystOpinions": info.get("numberOfAnalystOpinions"),
    }
    return json.dumps(price_targets, indent=2)


def register(mcp: FastMCP):
    """Register all public async callables in this module as tools."""
    module = sys.modules[__name__]
    for name, func in inspect.getmembers(module, inspect.isfunction):
        if name.startswith("_") or name == "register":
            continue
        mcp.tool()(func)



"""Market data and trending tools."""

import inspect
import json
import sys

import yfinance as yf
from mcp.server import FastMCP

from openmarkets.core.serializers import JSONSerializer


async def get_available_markets() -> list:
    """Get a list of available markets.

    Returns:
        List of market names
    """
    # Note: yfinance doesn't have a direct endpoint for available markets
    # This is a simplified implementation using major US indices
    return ["US", "GB", "ASIA", "EUROPE", "RATES", "COMMODITIES", "CURRENCIES", "CRYPTOCURRENCIES"]


async def get_market_status(market: str) -> str:
    """Get current market status and trading hours.

    Returns:
        JSON string containing market status information
    """
    # Note: yfinance doesn't have a direct market status endpoint
    # This is a simplified implementation using a major index
    try:
        market = yf.Market(market)
        status = market.status

        return json.dumps(status, indent=2, default=str, cls=JSONSerializer)
    except Exception as e:
        return json.dumps({"error": f"Failed to get market status: {str(e)}"})


async def get_market_summary(market: str) -> str:
    """Get current market summary information.

    Returns:
        JSON string containing market summary information
    """
    # Note: yfinance doesn't have a direct market summary endpoint
    # This is a simplified implementation using a major index
    try:
        market = yf.Market(market)
        summary = market.summary

        return json.dumps(summary, indent=2, default=str, cls=JSONSerializer)
    except Exception as e:
        return json.dumps({"error": f"Failed to get market summary: {str(e)}"})


async def get_sector_performance() -> str:
    """Get sector performance using sector ETFs.

    Returns:
        JSON string containing sector performance data
    """
    sector_etfs = {
        "Technology": "XLK",
        "Healthcare": "XLV",
        "Financials": "XLF",
        "Consumer Discretionary": "XLY",
        "Communication Services": "XLC",
        "Industrials": "XLI",
        "Consumer Staples": "XLP",
        "Energy": "XLE",
        "Utilities": "XLU",
        "Real Estate": "XLRE",
        "Materials": "XLB",
    }

    try:
        sector_performance = []

        for sector, etf in sector_etfs.items():
            stock = yf.Ticker(etf)
            info = stock.info
            hist = stock.history(period="2d")

            if len(hist) >= 2:
                daily_change = ((hist.iloc[-1]["Close"] - hist.iloc[-2]["Close"]) / hist.iloc[-2]["Close"]) * 100
            else:
                daily_change = None

            sector_performance.append(
                {
                    "sector": sector,
                    "etf_symbol": etf,
                    "current_price": info.get("currentPrice"),
                    "daily_change_percent": daily_change,
                    "volume": info.get("volume"),
                }
            )

        return json.dumps({"sector_performance": sector_performance}, indent=2)

    except Exception as e:
        return json.dumps({"error": f"Failed to get sector performance: {str(e)}"})


async def get_index_data(indices: list[str] = None) -> str:
    """Get major market indices data.

    Args:
        indices: List of index symbols (default: major US indices)

    Returns:
        JSON string containing index data
    """
    if indices is None:
        indices = ["^GSPC", "^DJI", "^IXIC", "^RUT", "^VIX"]

    index_names = {
        "^GSPC": "S&P 500",
        "^DJI": "Dow Jones Industrial Average",
        "^IXIC": "NASDAQ Composite",
        "^RUT": "Russell 2000",
        "^VIX": "CBOE Volatility Index",
    }

    try:
        index_data = []

        for index in indices:
            stock = yf.Ticker(index)
            hist = stock.history(period="2d")

            if len(hist) >= 2:
                current_price = hist.iloc[-1]["Close"]
                previous_close = hist.iloc[-2]["Close"]
                daily_change = current_price - previous_close
                daily_change_percent = (daily_change / previous_close) * 100
            else:
                current_price = None
                daily_change = None
                daily_change_percent = None

            index_data.append(
                {
                    "symbol": index,
                    "name": index_names.get(index, index),
                    "current_price": current_price,
                    "daily_change": daily_change,
                    "daily_change_percent": daily_change_percent,
                    "volume": hist.iloc[-1]["Volume"] if len(hist) > 0 else None,
                }
            )

        return json.dumps({"indices": index_data}, indent=2, cls=JSONSerializer)

    except Exception as e:
        return json.dumps({"error": f"Failed to get index data: {str(e)}"})


def register(mcp: FastMCP):
    """Register market data tools with the MCP server."""
    module = sys.modules[__name__]
    for name, func in inspect.getmembers(module, inspect.isfunction):
        if name.startswith("_") or name == "register":
            continue
        mcp.tool()(func)



import yfinance as yf

from openmarkets.schemas.analysis import (
    AnalystPriceTargets,
    AnalystRecommendation,
    AnalystRecommendationChange,
    EarningsEstimate,
    EPSTrend,
    GrowthEstimates,
    RevenueEstimate,
)


def fetch_analyst_recommendations(ticker: str) -> list[AnalystRecommendation]:
    """
    Fetch analyst recommendations for a given ticker.

    Args:
        ticker (str): The stock ticker symbol.

    Returns:
        list[AnalystRecommendation]: List of analyst recommendations.
    """
    df = yf.Ticker(ticker).get_recommendations()
    return [AnalystRecommendation(**row.to_dict()) for _, row in df.iterrows()]


def fetch_analyst_upgrades_downgrades(ticker: str) -> list[AnalystRecommendationChange]:
    """
    Fetch upgrades and downgrades for a given ticker.

    Args:
        ticker (str): The stock ticker symbol.

    Returns:
        list[AnalystRecommendationChange]: List of analyst recommendation changes.
    """
    df = yf.Ticker(ticker).get_upgrades_downgrades()
    df.reset_index(inplace=True)
    return [AnalystRecommendationChange(**row.to_dict()) for _, row in df.iterrows()]


def fetch_revenue_estimates(ticker: str) -> list[RevenueEstimate]:
    """
    Fetch revenue estimates for a given ticker.

    Args:
        ticker (str): The stock ticker symbol.

    Returns:
        list[RevenueEstimate]: List of revenue estimates.
    """
    df = yf.Ticker(ticker).get_revenue_estimate()
    df.reset_index(inplace=True)
    return [RevenueEstimate(**row.to_dict()) for _, row in df.iterrows()]


def fetch_earnings_estimates(ticker: str) -> list[EarningsEstimate]:
    """
    Fetch earnings estimates for a given ticker.

    Args:
        ticker (str): The stock ticker symbol.

    Returns:
        list[EarningsEstimate]: List of earnings estimates.
    """
    df = yf.Ticker(ticker).get_earnings_estimate()
    df.reset_index(inplace=True)
    return [EarningsEstimate(**row.to_dict()) for _, row in df.iterrows()]


def fetch_eps_trends(ticker: str) -> list[EPSTrend]:
    """
    Fetch EPS trends for a given ticker.

    Args:
        ticker (str): The stock ticker symbol.

    Returns:
        list[EPSTrend]: List of EPS trends.
    """
    df = yf.Ticker(ticker).get_eps_trend()
    df.reset_index(inplace=True)
    return [EPSTrend(**row.to_dict()) for _, row in df.iterrows()]


def fetch_growth_estimates(ticker: str) -> list[GrowthEstimates]:
    """
    Fetch growth estimates for a given ticker.

    Args:
        ticker (str): The stock ticker symbol.

    Returns:
        list[GrowthEstimates]: List of growth estimates.
    """
    df = yf.Ticker(ticker).get_growth_estimates()
    df.reset_index(inplace=True)
    return [GrowthEstimates(**row.to_dict()) for _, row in df.iterrows()]


def fetch_analyst_price_targets(ticker: str) -> AnalystPriceTargets:
    """
    Get analyst price targets and estimates for a given ticker.

    Args:
        ticker (str): The stock ticker symbol.

    Returns:
        AnalystPriceTargets: Analyst price targets and related estimates.
    """
    df = yf.Ticker(ticker).get_analyst_price_targets()
    return AnalystPriceTargets(**df)



import yfinance as yf

from openmarkets.schemas.markets import MarketStatus, MarketSummary, MarketType, SummaryEntry


def fetch_market_summary(market: str) -> MarketSummary:
    """
    Fetch the market summary for a given market type.

    Args:
        market (str): The market type.

    Returns:
        MarketSummary: Market summary data.
    """
    summary = yf.Market(market).summary
    return MarketSummary(summary={k: SummaryEntry(**v) for k, v in summary.items()})


def fetch_market_status(market: str) -> MarketStatus:
    """
    Fetch the market status for a given market type.

    Args:
        market (str): The market type.

    Returns:
        MarketStatus: Market status data.
    """
    status = yf.Market(market).status
    return MarketStatus(**status)



from typing import Any

from pydantic import BaseModel, Field

from .company import CompanyOfficer


class FundInfo(BaseModel):
    """Schema for fund information, typically from yfinance Ticker.info for funds/ETFs."""

    phone: str | None = None
    longBusinessSummary: str | None = None
    companyOfficers: list[CompanyOfficer] | None = None
    executiveTeam: list[Any] | None = None
    maxAge: int | None = None
    priceHint: int | None = None
    previousClose: float | None = None
    open: float | None = None
    dayLow: float | None = None
    dayHigh: float | None = None
    regularMarketPreviousClose: float | None = None
    regularMarketOpen: float | None = None
    regularMarketDayLow: float | None = None
    regularMarketDayHigh: float | None = None
    trailingPE: float | None = None
    volume: int | None = None
    regularMarketVolume: int | None = None
    averageVolume: int | None = None
    averageVolume10days: int | None = None
    averageDailyVolume10Day: int | None = None
    bid: float | None = None
    ask: float | None = None
    bidSize: int | None = None
    askSize: int | None = None
    yield_: float | None = Field(None, alias="yield", description="Fund yield.")
    totalAssets: float | None = None
    fiftyTwoWeekLow: float | None = None
    fiftyTwoWeekHigh: float | None = None
    allTimeHigh: float | None = None
    allTimeLow: float | None = None
    fiftyDayAverage: float | None = None
    twoHundredDayAverage: float | None = None
    trailingAnnualDividendRate: float | None = None
    trailingAnnualDividendYield: float | None = None
    navPrice: float | None = None
    currency: str | None = None
    tradeable: bool | None = None
    category: str | None = None
    ytdReturn: float | None = None
    beta3Year: float | None = None
    fundFamily: str | None = None
    fundInceptionDate: int | None = None
    legalType: str | None = None
    threeYearAverageReturn: float | None = None
    fiveYearAverageReturn: float | None = None
    quoteType: str | None = None
    symbol: str | None = None
    language: str | None = None
    region: str | None = None
    typeDisp: str | None = None
    quoteSourceName: str | None = None
    triggerable: bool | None = None
    customPriceAlertConfidence: str | None = None
    longName: str | None = None
    shortName: str | None = None
    marketState: str | None = None
    fiftyTwoWeekLowChangePercent: float | None = None
    fiftyTwoWeekRange: str | None = None
    fiftyTwoWeekHighChange: float | None = None
    fiftyTwoWeekHighChangePercent: float | None = None
    fiftyTwoWeekChangePercent: float | None = None
    dividendYield: float | None = None
    trailingThreeMonthReturns: float | None = None
    trailingThreeMonthNavReturns: float | None = None
    netAssets: float | None = None
    epsTrailingTwelveMonths: float | None = None
    bookValue: float | None = None
    fiftyDayAverageChange: float | None = None
    fiftyDayAverageChangePercent: float | None = None
    twoHundredDayAverageChange: float | None = None
    twoHundredDayAverageChangePercent: float | None = None
    netExpenseRatio: float | None = None
    priceToBook: float | None = None
    sourceInterval: int | None = None
    exchangeDataDelayedBy: int | None = None
    cryptoTradeable: bool | None = None
    hasPrePostMarketData: bool | None = None
    firstTradeDateMilliseconds: int | None = None
    postMarketChangePercent: float | None = None
    postMarketPrice: float | None = None
    postMarketChange: float | None = None
    regularMarketChange: float | None = None
    regularMarketDayRange: str | None = None
    fullExchangeName: str | None = None
    financialCurrency: str | None = None
    averageDailyVolume3Month: int | None = None
    fiftyTwoWeekLowChange: float | None = None
    corporateActions: list[Any] | None = None
    postMarketTime: int | None = None
    regularMarketTime: int | None = None
    exchange: str | None = None
    messageBoardId: str | None = None
    exchangeTimezoneName: str | None = None
    exchangeTimezoneShortName: str | None = None
    gmtOffSetMilliseconds: int | None = None
    market: str | None = None
    esgPopulated: bool | None = None
    regularMarketChangePercent: float | None = None
    regularMarketPrice: float | None = None
    trailingPegRatio: float | None = None


class FundEquityHolding(BaseModel):
    """Schema for individual equity holdings within a fund."""

    fund: str | None = Field(None, alias="index")
    price_to_earnings: float | None = Field(None, alias="Price/Earnings")
    price_to_book: float | None = Field(None, alias="Price/Book")
    price_to_sales: float | None = Field(None, alias="Price/Sales")
    price_to_cashflow: float | None = Field(None, alias="Price/Cashflow")
    median_market_cap: float | None = Field(None, alias="Median Market Cap")
    three_year_earnings_growth: float | None = Field(None, alias="3 Year Earnings Growth")

    model_config = {"arbitrary_types_allowed": True}


class FundHoldings(BaseModel):
    """Schema for fund holdings information."""

    equity_holdings: list[FundEquityHolding]
    total_equity_holdings: float | None = None
    total_fixed_income_holdings: float | None = None
    total_other_holdings: float | None = None
    total_holdings: float | None = None


class FundBondHolding(BaseModel):
    """Schema for individual bond holdings within a fund."""

    fund: str | None = Field(None, alias="index")
    duration: float | None = Field(None, alias="Duration")
    maturity: float | None = Field(None, alias="Maturity")
    credit_quality: float | None = Field(None, alias="Credit Quality")


class FundAssetClassHolding(BaseModel):
    """Schema for individual asset class holdings within a fund."""

    cashPosition: float | None = Field(None, description="Cash Position")
    stockPosition: float | None = Field(None, description="Stock Position")
    bondPosition: float | None = Field(None, description="Bond Position")
    preferredPosition: float | None = Field(None, description="Preferred Position")
    convertiblePosition: float | None = Field(None, description="Convertible Position")
    otherPosition: float | None = Field(None, description="Other Position")


class FundTopHolding(BaseModel):
    """Schema for top holdings within a fund."""

    Symbol: str = Field(..., description="Ticker symbol of the holding.")
    Name: str = Field(..., description="Name of the holding.")
    Holding_Percent: float = Field(
        ..., description="Percentage of the fund's total holdings represented by this holding.", alias="Holding Percent"
    )


class FundSectorWeighting(BaseModel):
    """Schema for sector weightings within a fund."""

    realestate: float | None = Field(None, description="Real Estate")
    customer_ciclical: float | None = Field(None, description="Consumer Cyclical")
    basic_materials: float | None = Field(None, description="Basic Materials")
    consumer_defensive: float | None = Field(None, description="Consumer Defensive")
    utilities: float | None = Field(None, description="Utilities")
    energy: float | None = Field(None, description="Energy")
    communication_services: float | None = Field(None, description="Communication Services")
    financial_services: float | None = Field(None, description="Financial Services")
    industrials: float | None = Field(None, description="Industrials")
    technology: float | None = Field(None, description="Technology")
    healthcare: float | None = Field(None, description="Healthcare")


class FundOperations(BaseModel):
    index: str | None = Field(None, description="Index or fund identifier.")
    annual_report_expense_ratio: float | None = Field(
        None, description="Annual report expense ratio of the fund.", alias="Annual Report Expense Ratio"
    )
    annual_holdings_turnover: float | None = Field(
        None, description="Annual holdings turnover of the fund.", alias="Annual Holdings Turnover"
    )
    total_net_assets: float | None = Field(None, description="Total net assets of the fund.", alias="Total Net Assets")


class FundOverview(BaseModel):
    categoryName: str | None = Field(None, description="Category name of the fund.")
    family: str | None = Field(None, description="Fund family.")
    legalType: str | None = Field(None, description="Legal type of the fund.")



import json

import numpy as np
import pandas as pd


class JSONSerializer(json.JSONEncoder):
    """
    Custom JSON encoder for pandas and numpy objects.

    Handles:
        - pandas.Timestamp (as ISO string)
        - numpy integer/floating types
        - numpy.ndarray, pandas.Series, pandas.DataFrame
        - NaN and None values
    """

    def default(self, o: object) -> object:
        if isinstance(o, pd.Timestamp):
            return o.isoformat()
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, (np.ndarray, pd.Series)):
            return o.tolist()
        if isinstance(o, pd.DataFrame):
            return o.to_dict(orient="records")
        if o is None or (isinstance(o, float) and np.isnan(o)):
            return None
        return super().default(o)


def safe_json_dumps(data: object, indent: int | None = None) -> str:
    """
    Serialize data to a JSON string, converting pandas/numpy objects to JSON serializable types.

    Args:
        data (object): The data to serialize.
        indent (Optional[int], optional): If not None, pretty-print with this indent level.

    Returns:
        str: JSON string.

    Raises:
        TypeError: If the data cannot be serialized.
    """
    try:
        return json.dumps(data, cls=JSONSerializer, indent=indent)
    except TypeError as exc:
        # Log or handle serialization errors as needed
        raise TypeError(f"Failed to serialize data: {exc}") from exc



# Application entry point (starts the server)

import logging
import sys

from openmarkets.core.server import main

logging.basicConfig(
    level=logging.INFO, stream=sys.stdout, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)

if __name__ == "__main__":
    main()



import pandas as pd
import yfinance as yf

from openmarkets.schemas.stock import (
    CorporateActions,
    NewsItem,
    StockDividends,
    StockFastInfo,
    StockHistory,
    StockInfo,
    StockSplit,
)


def fetch_fast_info(ticker: str) -> StockFastInfo:
    """
    Fetch fast stock info for a given ticker and return as StockFastInfo.
    """
    fast_info = yf.Ticker(ticker).fast_info
    return StockFastInfo(**fast_info)


def fetch_info(ticker: str) -> StockInfo:
    """
    Fetch detailed stock info for a given ticker and return as StockInfo.
    """
    info = yf.Ticker(ticker).info
    return StockInfo(**info)


def fetch_history(ticker: str, period: str = "1y", interval: str = "1d") -> list[StockHistory]:
    """
    Fetch historical OHLCV data for a given ticker and return as a list of StockHistory.
    """
    df = yf.Ticker(ticker).history(period=period, interval=interval)
    df.reset_index(inplace=True)
    return [StockHistory(**row.to_dict()) for _, row in df.iterrows()]


def fetch_dividends(ticker: str) -> list[StockDividends]:
    """
    Fetch dividend history for a given ticker and return as a list of StockDividends.
    """
    dividends = yf.Ticker(ticker).dividends
    return [StockDividends(Date=row[0], Dividends=row[1]) for row in dividends.to_dict().items()]


def fetch_financial_summary(ticker: str) -> dict:
    """
    Fetch financial summary data for a given ticker and return as a dictionary.
    """
    include_fields = {
        "totalRevenue",
        "revenueGrowth",
        "grossProfits",
        "grossMargins",
        "operatingMargins",
        "profitMargins",
        "operatingCashflow",
        "freeCashflow",
        "totalCash",
        "totalDebt",
        "totalCashPerShare",
        "earningsGrowth",
        "currentRatio",
        "quickRatio",
        "returnOnAssets",
        "returnOnEquity",
        "debtToEquity",
    }
    data = yf.Ticker(ticker).info
    return StockInfo(**data).model_dump(include=include_fields)


def fetch_risk_metrics(ticker: str) -> dict:
    """
    Fetch risk metrics data for a given ticker and return as a dictionary.
    """
    include_fields = {
        "auditRisk",
        "boardRisk",
        "compensationRisk",
        "financialRisk",
        "governanceRisk",
        "overallRisk",
        "shareHolderRightsRisk",
    }
    data = yf.Ticker(ticker).info
    return StockInfo(**data).model_dump(include=include_fields)


def fetch_dividend_summary(ticker: str) -> dict:
    """
    Fetch dividend summary data for a given ticker and return as a dictionary.
    """
    include_fields = {
        "dividendRate",
        "dividendYield",
        "payoutRatio",
        "fiveYearAvgDividendYield",
        "trailingAnnualDividendRate",
        "trailingAnnualDividendYield",
        "exDividendDate",
        "lastDividendDate",
        "lastDividendValue",
    }
    data = yf.Ticker(ticker).info
    return StockInfo(**data).model_dump(include=include_fields)


def fetch_price_target(ticker: str) -> dict:
    """
    Fetch analyst price target data for a given ticker and return as a dictionary.
    """
    include_fields = {
        "targetHighPrice",
        "targetLowPrice",
        "targetMeanPrice",
        "targetMedianPrice",
        "recommendationMean",
        "recommendationKey",
        "numberOfAnalystOpinions",
    }
    data = yf.Ticker(ticker).info
    return StockInfo(**data).model_dump(include=include_fields)


def fetch_financial_summary_v2(ticker: str) -> dict:
    """
    Fetch financial summary data for a given ticker and return as a dictionary.
    """
    include_fields = {
        "marketCap",
        "enterpriseValue",
        "floatShares",
        "sharesOutstanding",
        "sharesShort",
        "bookValue",
        "priceToBook",
        "totalRevenue",
        "revenueGrowth",
        "grossProfits",
        "grossMargins",
        "operatingMargins",
        "profitMargins",
        "operatingCashflow",
        "freeCashflow",
        "totalCash",
        "totalDebt",
        "totalCashPerShare",
        "earningsGrowth",
        "currentRatio",
        "quickRatio",
        "returnOnAssets",
        "returnOnEquity",
        "debtToEquity",
    }
    data = yf.Ticker(ticker).info
    return StockInfo(**data).model_dump(include=include_fields)


def fetch_quick_technical_indicators(ticker: str) -> dict:
    """
    Fetch technical indicators for a given ticker and return as a dictionary.
    """
    include_fields = {
        "currentPrice",
        "fiftyDayAverage",
        "twoHundredDayAverage",
        "fiftyDayAverageChange",
        "fiftyDayAverageChangePercent",
        "twoHundredDayAverageChange",
        "twoHundredDayAverageChangePercent",
        "fiftyTwoWeekLow",
        "fiftyTwoWeekHigh",
    }
    data = yf.Ticker(ticker).info
    return StockInfo(**data).model_dump(include=include_fields)


def fetch_splits(ticker: str) -> list[StockSplit]:
    """
    Fetch stock split history for a given ticker and return as a list of StockSplit.
    """
    splits = yf.Ticker(ticker).splits
    return [
        StockSplit(date=pd.Timestamp(str(index)).to_pydatetime(), stock_splits=value) for index, value in splits.items()
    ]


def fetch_corporate_actions(ticker: str) -> list[CorporateActions]:
    """
    Fetch corporate actions (splits/dividends) history for a given ticker and return as a list of CorporateActions.
    """
    actions = yf.Ticker(ticker).actions
    return [CorporateActions(**row.to_dict()) for _, row in actions.reset_index().iterrows()]


def fetch_news(ticker: str) -> list[NewsItem]:
    """
    Fetch news items for a given ticker and return as a list of NewsItem.
    """
    news = yf.Ticker(ticker).news
    return [NewsItem(**item) for item in news]
